@@ -14,7 +14,9 @@ from medcat.preprocessing.cleaners import spacy_tag_punct
 from medcat.utils.helpers import get_all_from_name, tkn_inds_from_doc
 from medcat.utils.loggers import basic_logger
 from medcat.utils.data_utils import make_mc_train_test
+import time
 import sys, traceback
+from tqdm.autonotebook import tqdm
 
 log = basic_logger("CAT")
 
@@ -50,7 +52,7 @@ class CAT(object):
         >>>spacy_doc = cat("Put some text here")
         >>>print(spacy_doc.ents) # Detected entites
     '''
-    def __init__(self, cdb, vocab=None, skip_stopwords=True, meta_cats=[], config={}):
+    def __init__(self, cdb, vocab=None, skip_stopwords=True, meta_cats=[], config={}, tokenizer=None):
         self.cdb = cdb
         self.vocab = vocab
         self.config = config
@@ -68,7 +70,7 @@ class CAT(object):
         self.nlp.add_spell_checker(spell_checker=self.spell_checker)
 
         # Add them cat class that does entity detection
-        self.spacy_cat = SpacyCat(cdb=self.cdb, vocab=self.vocab)
+        self.spacy_cat = SpacyCat(cdb=self.cdb, vocab=self.vocab, tokenizer=tokenizer)
         self.nlp.add_cat(spacy_cat=self.spacy_cat)
 
         # Add meta_annotaiton classes if they exist
@@ -335,6 +337,7 @@ class CAT(object):
         cui_rec = {}
         cui_f1 = {}
         cui_counts = {}
+        examples = {'fp': {}, 'fn': {}, 'tp': {}}
 
         fp_docs = set()
         fn_docs = set()
@@ -347,21 +350,21 @@ class CAT(object):
         else:
             _cui_filter = list(self.spacy_cat.CUI_FILTER)
 
-        # Stupid
-        for project in data['projects']:
+        for pind, project in tqdm(enumerate(data['projects']), desc="Stats project", total=len(data['projects']), leave=False):
             cui_filter = None
             tui_filter = None
 
             if use_filters:
                 if 'cuis' in project and len(project['cuis'].strip()) > 0:
-                    cui_filter = [x.strip() for x in project['cuis'].split(",")]
+                    cui_filter = set([x.strip() for x in project['cuis'].split(",")])
                 if 'tuis' in project and len(project['tuis'].strip()) > 0:
-                    tui_filter = [x.strip().upper() for x in project['tuis'].split(",")]
+                    tui_filter = set([x.strip().upper() for x in project['tuis'].split(",")])
 
                 self.spacy_cat.TUI_FILTER = tui_filter
                 self.spacy_cat.CUI_FILTER = cui_filter
 
-            for doc in project['documents']:
+            start_time = time.time()
+            for dind, doc in tqdm(enumerate(project['documents']), desc='Stats document', total=len(project['documents']), leave=False):
                 spacy_doc = self(doc['text'])
                 anns = doc['annotations']
                 if use_overlaps:
@@ -370,6 +373,8 @@ class CAT(object):
                     p_anns = spacy_doc.ents
 
                 anns_norm = []
+                anns_norm_neg = []
+                anns_examples = []
                 anns_norm_cui = []
                 for ann in anns:
                     if (cui_filter is None and tui_filter is None) or (cui_filter is not None and ann['cui'] in cui_filter) or \
@@ -380,6 +385,15 @@ class CAT(object):
 
                         if ann.get('validated', True) and (not ann.get('killed', False) and not ann.get('deleted', False)):
                             anns_norm.append((ann['start'], cui))
+                            anns_examples.append({"text": doc['text'][max(0, ann['start']-60):ann['end']+60],
+                                                  "cui": cui,
+                                                  "source value": ann['value'],
+                                                  "acc": 1,
+                                                  "project index": pind,
+                                                  "document inedex": dind})
+                        elif ann.get('validated', True) and (ann.get('killed', False) or ann.get('deleted', False)):
+                            anns_norm_neg.append((ann['start'], cui))
+
 
                         if ann.get("validated", True):
                             # This is used to test was someone annotating for this CUI in this document
@@ -387,30 +401,51 @@ class CAT(object):
                             cui_counts[cui] = cui_counts.get(cui, 0) + 1
 
                 p_anns_norm = []
+                p_anns_examples = []
                 for ann in p_anns:
                     cui = ann._.cui
                     if use_groups:
                         cui = self.cdb.cui2info.get(cui, {}).get("group", cui)
                     p_anns_norm.append((ann.start_char, cui))
+                    p_anns_examples.append({"text": doc['text'][max(0, ann.start_char-60):ann.end_char+60],
+                                          "cui": cui,
+                                          "source value": ann.text,
+                                          "acc": float(ann._.acc),
+                                          "project index": pind,
+                                          "document inedex": dind})
 
-                for ann in p_anns_norm:
+
+                for iann, ann in enumerate(p_anns_norm):
                     if not use_cui_doc_limit or ann[1] in anns_norm_cui:
                         cui = ann[1]
                         if ann in anns_norm:
                             tp += 1
                             tps[cui] = tps.get(cui, 0) + 1
+
+                            example = p_anns_examples[iann]
+                            examples['tp'][cui] = examples['tp'].get(cui, []) + [example]
                         else:
                             fp += 1
                             fps[cui] = fps.get(cui, 0) + 1
                             fp_docs.add(doc['name'])
 
-                for ann in anns_norm:
+                            # Add example for this FP prediction
+                            example = p_anns_examples[iann]
+                            if ann in anns_norm_neg:
+                                # Means that it really was annotated as negative
+                                example['real_fp'] = True
+
+                            examples['fp'][cui] = examples['fp'].get(cui, []) + [example]
+
+                for iann, ann in enumerate(anns_norm):
                     if ann not in p_anns_norm:
                         cui = ann[1]
                         fn += 1
                         fn_docs.add(doc['name'])
 
                         fns[cui] = fns.get(cui, 0) + 1
+                        examples['fn'][cui] = examples['fn'].get(cui, []) + [anns_examples[iann]]
+
         try:
             prec = tp / (tp + fp)
             rec = tp / (tp + fn)
@@ -462,12 +497,13 @@ class CAT(object):
         self.spacy_cat.TUI_FILTER = _tui_filter
         self.spacy_cat.CUI_FILTER = _cui_filter
 
-        return fps, fns, tps, cui_prec, cui_rec, cui_f1, cui_counts
+        return fps, fns, tps, cui_prec, cui_rec, cui_f1, cui_counts, examples
 
 
     def train_supervised(self, data_path, reset_cdb=False, reset_cui_count=False, nepochs=30, lr=None,
                          anneal=None, print_stats=True, use_filters=False, terminate_last=False, use_overlaps=False,
-                         use_cui_doc_limit=False, test_size=0, force_manually_created=False, use_groups=False):
+                         use_cui_doc_limit=False, test_size=0, force_manually_created=False, use_groups=False,
+                         never_terminate=False):
         r'''
         Run supervised training on a dataset from MedCATtrainer. Please take care that this is more a simiulated
         online training then supervised.
@@ -511,6 +547,8 @@ class CAT(object):
                 created.
             use_groups (boolean):
                 If True concepts that have groups will be combined and stats will be reported on groups.
+            never_terminate (boolean):
+                If True no termination will be applied
 
         Returns:
             fp (dict):
@@ -527,8 +565,10 @@ class CAT(object):
                 F1 for each CUI
             cui_counts (dict):
                 Number of occurrence for each CUI
+            examples (dict):
+                FP/FN examples of sentences for each CUI
         '''
-        fp = fn = tp = p = r = f1 = cui_counts = {}
+        fp = fn = tp = p = r = f1 = cui_counts = examples = {}
 
         self.train = False
         data = json.load(open(data_path))
@@ -561,19 +601,17 @@ class CAT(object):
                     self.cdb.cui_count[cui] = 10
 
         # Remove entites that were terminated
-        for project in train_set['projects']:
-            for doc in project['documents']:
-                for ann in doc['annotations']:
-                    if ann.get('killed', False):
-                        self.unlink_concept_name(ann['cui'], ann['value'])
-
-        for epoch in range(nepochs):
-            print("Starting epoch: {}".format(epoch))
-            log.info("Starting epoch: {}".format(epoch))
-            # Print acc before training
-
+        if not never_terminate:
             for project in train_set['projects']:
-                for i_doc, doc in enumerate(project['documents']):
+                for doc in project['documents']:
+                    for ann in doc['annotations']:
+                        if ann.get('killed', False):
+                            self.unlink_concept_name(ann['cui'], ann['value'])
+
+        for epoch in tqdm(range(nepochs), desc='Epoch', leave=False):
+            # Print acc before training
+            for project in tqdm(train_set['projects'], desc='Project', leave=False, total=len(train_set['projects'])):
+                for i_doc, doc in tqdm(enumerate(project['documents']), desc='Document', leave=False, total=len(project['documents'])):
                     spacy_doc = self(doc['text'])
                     for ann in doc['annotations']:
                         if not ann.get('killed', False):
@@ -592,7 +630,7 @@ class CAT(object):
                                           lr=lr,
                                           anneal=anneal,
                                           manually_created=manually_created)
-            if terminate_last:
+            if terminate_last and not never_terminate:
                 # Remove entites that were terminated, but after all training is done
                 for project in train_set['projects']:
                     for doc in project['documents']:
@@ -602,12 +640,12 @@ class CAT(object):
 
             if epoch % 5 == 0:
                 if print_stats:
-                    fp, fn, tp, p, r, f1, cui_counts = self._print_stats(test_set, epoch=epoch+1,
+                    fp, fn, tp, p, r, f1, cui_counts, examples = self._print_stats(test_set, epoch=epoch+1,
                                                              use_filters=use_filters,
                                                              use_cui_doc_limit=use_cui_doc_limit,
                                                              use_overlaps=use_overlaps,
                                                              use_groups=use_groups)
-        return fp, fn, tp, p, r, f1, cui_counts
+        return fp, fn, tp, p, r, f1, cui_counts, examples
 
     @property
     def train(self):
