@@ -3,8 +3,10 @@ import numpy as np
 import operator
 from medcat.utils.loggers import basic_logger
 from medcat.utils.matutils import unitvec
+from medcat.utils.ml_utils import load_hf_tokenizer, build_vocab_from_hf
+from spacy.lang.en.stop_words import STOP_WORDS
 import os
-log = basic_logger("cat_spacycat")
+log = basic_logger("spacycat")
 
 # IF UMLS it includes specific rules that are only good for the Full UMLS version
 if os.getenv('TYPE', 'other').lower() == 'umls':
@@ -28,7 +30,7 @@ class SpacyCat(object):
     DEBUG = os.getenv('DEBUG', "false").lower() == 'true'
     NORM_EMB = os.getenv('NORM_EMB', "false").lower() == 'true' # Should we normalize the w2v
     PREFER_FREQUENT = os.getenv('PREFER_FREQUENT', "false").lower() == 'true'
-    PREFER_ICD10 = os.getenv('PREFER_ICD10', "false").lower() == 'true'
+    PREFER_CONCEPTS_WITH = os.getenv('PREFER_CONCEPTS_WITH', None)
     CNTX_SPAN = int(os.getenv('CNTX_SPAN', 9))
     CNTX_SPAN_SHORT = int(os.getenv('CNTX_SPAN_SHORT', 3))
     MIN_CUI_COUNT = int(os.getenv('MIN_CUI_COUNT', 30000))
@@ -69,9 +71,14 @@ class SpacyCat(object):
 
         if self.vocab is None:
             self.vocab = self.cdb.vocab
-
         if tokenizer is None:
             self.tokenizer = self._tok
+        elif type(tokenizer) == str:
+            self.hf_tokenizer = load_hf_tokenizer(tokenizer_name=tokenizer)
+            self.tokenizer = self._tok_hf
+
+            # Build hf vocab from BERT/Transofmer models
+            build_vocab_from_hf(model_name=tokenizer, hf_tokenizer=self.hf_tokenizer, vocab=self.vocab)
         else:
             self.tokenizer = tokenizer
 
@@ -79,7 +86,16 @@ class SpacyCat(object):
         self.wdrops = [1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2] + [0.1] * 300
 
 
-    def _tok(self, text):
+    def _tok_hf(self, token):
+        text = token.text
+        return self.hf_tokenizer.tokenize(text)
+
+    def _tok(self, token):
+        if self.NORM_EMB:
+            text = token._.norm
+        else:
+            text = token.text.lower()
+
         return [text]
 
 
@@ -101,37 +117,39 @@ class SpacyCat(object):
         add_weight = True
         while(n < span and i >= 0):
             word = doc[i]
-            if self.NORM_EMB:
-                _word = word._.norm
-            else:
-                _word = word.text.lower()
-
-            if self.WEIGHTED_AVG:
-                if add_weight:
-                    weights = [self.wdrops[n]] + weights
-                    add_weight = False
-            else:
-                weights.append(1.0)
+            new_words = []
 
             if skip_words:
                 if not word._.to_skip and not word.is_digit:
-                    words = self.tokenizer(_word) + words
+                    new_words = self.tokenizer(word)
+                    words = new_words + words
 
                     n += 1
                     add_weight = True
             else:
-                words = self.tokenizer(_word) + words
+                new_words = self.tokenizer(word)
+                words = new_words + words
 
                 n += 1
                 add_weight = True
 
-            i = i - 1
+            if self.WEIGHTED_AVG:
+                if add_weight:
+                    nwords = len(new_words)
+                    weights = [self.wdrops[n]] * nwords + weights
+                    add_weight = False
+            else:
+                nwords = len(new_words)
+                weights = weights + [1.0] * nwords
 
+
+            i = i - 1
         # Add tokens if not skip_current
         if not skip_current:
             for tkn in tkns:
-                words.append(tkn._.norm)
-                weights.append(1.0)
+                new_words = self.tokenizer(tkn)
+                words = words + new_words
+                weights = weights + [1.0] * len(new_words)
 
         # Go right
         i = tkns[-1].i + 1
@@ -139,28 +157,31 @@ class SpacyCat(object):
         add_weight = True
         while(n < span and i < len(doc)):
             word = doc[i]
-            if self.NORM_EMB:
-                _word = word._.norm
-            else:
-                _word = word.text.lower()
-
-            if self.WEIGHTED_AVG:
-                if add_weight:
-                    weights.append(self.wdrops[n])
-                    add_weight = False
-            else:
-                weights.append(1.0)
+            new_words = []
 
             if skip_words:
                 if not word._.to_skip and not word.is_digit:
-                    words = words + self.tokenizer(_word)
+                    new_words = self.tokenizer(word)
+                    words = words + new_words
                     n += 1
                     add_weight = True
             else:
-                words = words + self.tokenizer(_word)
+                new_words = self.tokenizer(word)
+                words = words + new_words
+
                 n += 1
                 add_weight = True
 
+            if self.WEIGHTED_AVG:
+                if add_weight:
+                    nwords = len(new_words)
+                    weights = weights + [self.wdrops[n]] * nwords
+                    add_weight = False
+            else:
+                nwords = len(new_words)
+                weights = weights + [1.0] * nwords
+
+
             i = i + 1
 
         return words, weights
@@ -287,8 +308,9 @@ class SpacyCat(object):
                     negative=negative, lr=lr, anneal=anneal)
 
         if np.random.rand() < self.NEG_PROB and not negative:
+            log.debug("Adding NEGATIVE for: " + str(cui))
             # Add only if probability and 'not' negative input
-            negs = self.vocab.get_negative_samples(n=self.CNTX_SPAN * 2)
+            negs = self.vocab.get_negative_samples(n=self.CNTX_SPAN * 2, ignore_punct_and_num=True, stopwords=STOP_WORDS)
             neg_cntx_vecs = [self.vocab.vec(self.vocab.index2word[x]) for x in negs]
             neg_cntx = np.average(neg_cntx_vecs, axis=0)
             self.cdb.add_context_vec(cui, neg_cntx, negative=True, cntx_type='MED',
@@ -559,7 +581,6 @@ class SpacyCat(object):
                         # If not just set the accuracy to -1
                         accs.append(-1)
                         cnts.append(-1)
-
                 # TODO: SEPARATE INTO A CLASS THAT MODIFIES THE SCORES
                 if self.PREFER_FREQUENT and cnts and max(cnts) > 100:
                     # Prefer frequent concepts, only in cases when cnt > 100
@@ -571,14 +592,13 @@ class SpacyCat(object):
 
                     accs = accs * mps
 
-                if self.PREFER_ICD10:
+                if self.PREFER_CONCEPTS_WITH is not None:
                     # Prefer concepts that have ICD10
                     mps = np.array([1] * len(cnts), dtype=np.float)
-                    noicd = [False if 'icd10' in self.cdb.cui2info.get(cui, {}) else
+                    noicd = [False if self.PREFER_CONCEPTS_WITH in self.cdb.cui2info.get(cui, {}) else
                              True for cui in cuis]
-                    mps[noicd] = 0.5
+                    mps[noicd] = 0.8
                     accs = accs * mps
-
                 ind = np.argmax(accs)
                 cui = cuis[ind]
                 acc = accs[ind]
